#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#define min(x,y) ( (x<y) ? x : y )
#define interp(val,newval,left,counts) val = ((left>0) ? ((val*left+newval*(counts-left))/counts) : newval)

#define PI 3.14159
#include "interface_2001.h"
#define FRAME_DELAY 500000

int light_data[60];

int trailing_rings(int count, int dir, int rot_dir, int ring_speed, int diff);
int sine_wave(int count,float speed, float speed2, float diff, int tran_counts);
int sine_spin(int count, int tran_counts, float speed, float diff, float off);

int count_usecs(struct timeval *tv);

int main()
{
	int ret;
	
	server_port=7879;
	if (init_lights()) {
		fprintf(stderr,"init_lights failed!\n");
		return -1;
	}

	srand(time(NULL));
	
/*	rise(100,5,10,3,0,-40,-80);

	rainfall(100,4,0);

	wipe(5,10);
*/	

	for (;;) {
		simple_count_pattern();	
	}

	for (;;) {
	trailing_lines(140,1,0,1);
	trailing_lines(140,-1,0,1);

	sine_spin(100,30,7.0,1,3.3);
	sine_spin(100,30,-7.0,1,3.3);

	sine_wave(200,-30.0,30.0,.5,30);

	wipe(5,10);

	sine_wave(200,-30.0,-30.0,.5,10);
	sine_wave(200,-30.0,30.0,.5,10);
	
	trailing_rings(160,1,0,2,0);
	trailing_rings(160,-1,0,2,0);

	sine_wave(200,-30.0,30.0,.5,30);
	sine_wave(200,-30.0,-30.0,.5,10);
	
	trailing_lines(240,1,0,1);
	trailing_lines(240,-1,0,1);
	}
	deinit_lights();
}

int count_usecs(struct timeval *tv)
{
	int diff;
	struct timeval mtv;
	
	gettimeofday(&mtv,NULL);
	diff=(mtv.tv_sec-tv->tv_sec)*1000000;
	diff+=mtv.tv_usec-tv->tv_usec;
	memcpy(tv,&mtv,sizeof(struct timeval));
	return diff;
}

int simple_count_pattern()
{
	int c,i;
	
	for (c=0;c<60;c++) {
//		for (i=0;i<60;i++) if (light_data[i]>0) light_data[i]-=60;
		for (i=0;i<60;i++) light_data[i]=0;
		light_data[c]=255;
		
		if (renderFrame(light_data)) return 1;
		usleep(FRAME_DELAY);
	}	
	return 0;
}

int wipe(int center, int speed)
{
	int lc;
	int x,y;
	int flag=1;
	
	
	while(flag==1) {
		flag=0;
		for (x=0;x<6;x++) {
			for (y=0;y<center;y++) {
				if (light_data[(y+1)*6+x]>0) {
					flag=1;
					light_data[y*6+x]+=min(light_data[(y+1)*6+x],speed);
					light_data[(y+1)*6+x]-=min(light_data[(y+1)*6+x],speed);
					if (light_data[y*6+x]>255) light_data[y*6+x]=255;
				}
			}
			
			for (y=(NUM_SEGMENTS-1);y>center;y--) {
				if (light_data[(y-1)*6+x]>0) {
					flag=1;
					light_data[y*6+x]+=min(light_data[(y-1)*6+x],speed);
					light_data[(y-1)*6+x]-=min(light_data[(y-1)*6+x],speed);
					if (light_data[y*6+x]>255) light_data[y*6+x]=255;
				}
			
			}
			
		}
		
		renderFrame(light_data);
	}
	return 0;
}
		
#define R_MAX_LINES 200
#define R_FADE_COUNT 20
#define R_NEW_LINE_CHANCE 3

// like trailing lines but falling from the ceiling...
int rainfall(int count, int speed, int die)
{
	int line_position[R_MAX_LINES];
	int line_count[R_MAX_LINES];
	int num_lines=1;
	int i;
	int loop_counter;
	
	for (i=0;i<R_MAX_LINES;i++) {
		line_position[i]=-1;
		line_count[i]=-1;
	}
	
//	bzero(light_data,60*sizeof(int));
	
	line_position[0]=(rand()%NUM_SEGMENTS)*6+2+rand()%2;
	line_count[0]=speed;
	light_data[line_position[0]]=255;
	
	for (loop_counter=0;loop_counter<count;loop_counter++) {
		
		// first go through and decrement all the existing trails
		
		for(i=0;i<60;i++) {
			light_data[i]-=R_FADE_COUNT;
			if (light_data[i]<0) light_data[i]=0;
		}
		
		// now go through and move all the lines that might need to be
		
		for (i=0;i<R_MAX_LINES;i++) {
			if (line_position[i]>=0) {
				line_count[i]--;
				if (line_count[i]==0) {
					line_count[i]=speed;
					if ((line_position[i]%6)>2) {
						line_position[i]++;
						if ((line_position[i]%6)==0) {
							if (die) line_position[i]=-1;
								else line_position[i]--;
						}
					} else {
						line_position[i]--;
						if ((line_position[i]%6)==5) {
							if (die) line_position[i]=-1;
								line_position[i]++;
						}
					}
					if (line_position[i]==-1) num_lines--;
				}
			}
		}
		
		if ( (( (rand()%R_NEW_LINE_CHANCE)==0)  && (num_lines<R_MAX_LINES)) || (num_lines==0)) {
			// create a new line
			
			for (i=0;i<R_MAX_LINES;i++) {
				if (line_position[i]==-1) break;
			}
			
			line_position[i]=(rand()%NUM_SEGMENTS)*6+2+rand()%2;
			line_count[i]=speed;
		}


		// set all the line heads to 255
		for (i=0;i<R_MAX_LINES;i++) {
			if (line_position[i]>=0) {
				light_data[line_position[i]]=255;
			}
		}
		
		
		renderFrame(light_data);
		usleep(FRAME_DELAY);
	}
	
	return 0;	
}
	

int pong(int count)
{
	int lc;
	int p1pos;
	int p2pos;

	int bx,by;	
	int bxs;
	int bys;
	
	for (lc=0;lc<count;lc++) {
		
		renderFrame(light_data);
	}
}

int sine_spin(int count, int tran_counts, float speed, float diff, float off)
{
	float val;
	float t;
	int lc;
	int x,y;
	int tran_left=tran_counts;
	
	for (lc=0;lc<count;lc++,tran_left--) {
		for (x=0;x<6;x++) {
			
			for (y=0;y<NUM_SEGMENTS;y++) {
    			t=(float) x+ (float)lc/speed + (float)y/off;
    			val=fabs(sin( PI*(float)t/6.0))*diff+(1.0-diff);
    			
				interp(light_data[y*6+x],(int) (val*255.0), tran_left, tran_counts);
				// interp might be the problem here
			}
		}			
	
		renderFrame(light_data);
		usleep(FRAME_DELAY);	
	}
	return 0;
}

int rise(int count, int center, int diff, int speed, int bot, int mid, int top)
{
	int lc;
	int x,y;
	int seg_level[NUM_SEGMENTS];
	
	seg_level[center]=0;
	for(y=(center+1);y<NUM_SEGMENTS;y++) seg_level[y]=seg_level[y-1]-diff;
	for(y=(center-1);y>=0;y--) seg_level[y]=seg_level[y+1]-diff;
	
	
	memset(light_data,0,sizeof(int)*NUM_SEGMENTS*6);
	
	for (lc=0;lc<count;lc++) {

		
		for (y=0;y<NUM_SEGMENTS;y++) {
			seg_level[y]+=speed;
			
			if ((seg_level[y]+bot)>0) 
				light_data[y*6]=light_data[y*6+5]=min(seg_level[y]+bot,255);
			if ((seg_level[y]+mid)>0) 
				light_data[y*6+1]=light_data[y*6+4]=min(seg_level[y]+mid,255);
			if ((seg_level[y]+top)>0) 
				light_data[y*6+2]=light_data[y*6+3]=min(seg_level[y]+top,255);
		
		}
		
		renderFrame(light_data);
		usleep(FRAME_DELAY);
	}
	return 0;
}
int sine_wave(int count,float speed, float speed2, float diff, int tran_counts)
{
	int lc,y,x,t;
	float val;
	int tran_left=tran_counts;
	
	for (lc=0;lc<count;lc++,tran_left--) {
		
		for (y=0;y<NUM_SEGMENTS;y++) {
			val=fabs(sin(((float)y/2.0)+(float)lc/speed))*255.0;
			t=(int) val;
			for(x=0;x<6;x++) {
				val=(fabs(cos((float)x/(6.0*PI)+(float)lc/speed2))*diff+(1-diff))*(float)t;
				t=(int) val;
				if (tran_left>0) light_data[y*6+x]=(light_data[y*6+x]*tran_left+t*(tran_counts-tran_left))/tran_counts;
					else light_data[y*6+x]=t;
			}
		}
		
		renderFrame(light_data);
		usleep(FRAME_DELAY);
	}
	return 0;
}
			

#define TR_FADE_COUNT 10

int trailing_rings(int count, int dir, int rot_dir, int ring_speed, int diff)
{
	int ring_position;
	int ring_count;
	int ring_accel;
	int loop_counter,i;
	int ring[]={255,255,255,255,255,255};
	int ring_index=0;

	ring[0]-=2*diff;
	ring[6]-=2*diff;
	ring[1]-=diff;
	ring[2]-=diff;

	ring_position=(dir>0) ? 0 : (NUM_SEGMENTS-1);
	ring_count=ring_speed;
	ring_accel=0;
	
	for (loop_counter=0;loop_counter<count;loop_counter++) {
		// first go and decrement all existing spaces
		for(i=0;i<(NUM_SEGMENTS*6);i++) {
			light_data[i]-=TR_FADE_COUNT;
			if (light_data[i]<0) light_data[i]=0;
		}
		
		ring_count--;
		if (ring_count==0) {
			ring_position+=dir;
			ring_index=(ring_index+rot_dir+6)%6;
			ring_count=ring_speed;
			if ((ring_position>=NUM_SEGMENTS) || (ring_position<0)) 
				ring_position=(dir>0) ? 0 : (NUM_SEGMENTS-1);
		}
		
		for (i=0;i<6;i++) light_data[ring_position*6+i]=ring[(ring_index+i)%6];
		
		renderFrame(light_data);
		
		usleep(FRAME_DELAY);
	}
	return 0;
}
			
	
	

#define TL_MAX_LINES 12
#define TL_FADE_COUNT 20
#define TL_NEW_LINE_CHANCE 3
	
int trailing_lines(int count, int dir, int rot_dir, int move_count)
{
	int line_position[TL_MAX_LINES];
	int line_count[TL_MAX_LINES];
	int num_lines=1;
	int i;
	int loop_counter;
	
	for (i=0;i<TL_MAX_LINES;i++) {
		line_position[i]=-1;
		line_count[i]=-1;
	}
	
//	bzero(light_data,60*sizeof(int));
	
	line_position[0]=rand()%6 + ( (dir>0) ? 0 : (NUM_SEGMENTS-1)*6);
	line_count[0]=move_count;
	light_data[line_position[0]]=255;
	
	for (loop_counter=0;loop_counter<count;loop_counter++) {
		
		// first go through and decrement all the existing trails
		
		for(i=0;i<60;i++) {
			light_data[i]-=TL_FADE_COUNT;
			if (light_data[i]<0) light_data[i]=0;
		}
		
		// now go through and move all the lines that might need to be
		
		for (i=0;i<TL_MAX_LINES;i++) {
			if (line_count[i]>0) {
				line_count[i]--;
				if (line_count[i]==0) {
					line_count[i]=move_count;
					line_position[i]+=(6*dir)+rot_dir;
//					if ((rand()%3)==0) 	line_position[i]+=rand()%3-1;
				}
			}
		}
		
		// check for out of range line positions
		for (i=0;i<TL_MAX_LINES;i++) {
			if ((line_position[i]>=(NUM_SEGMENTS*6)) || (line_position[i]<0)) {
				line_position[i]=-1;
				line_count[i]=-1;
				num_lines--;
			}
		}
		
		if ( (( (rand()%TL_NEW_LINE_CHANCE)==0)  && (num_lines<TL_MAX_LINES)) || (num_lines==0)) {
			// create a new line
			
			for (i=0;i<TL_MAX_LINES;i++) {
				if (line_position[i]==-1) break;
			}
			
			line_position[i]=rand()%6 + ( (dir>0) ? 0 : (NUM_SEGMENTS-1)*6);
			line_count[i]=move_count;
		}


		// set all the line heads to 255
		for (i=0;i<TL_MAX_LINES;i++) {
			if (line_position[i]>=0) {
				light_data[line_position[i]]=255;
			}
		}
		
		
		renderFrame(light_data);
		usleep(FRAME_DELAY);
	}
	
	return 0;	
}	
			
		