#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>

/* Receives data for the lights over a configurable UDP port and writes
   it out to the light interface, be it the real lights or a simulation. */

#define SERVER_IP "10.0.0.13"
#define SERVER_PORT 7878
   
#include "interface_2001.h"

int sock;
int light_data[NUM_SEGMENTS*6];

int main(int argc, char **argv)
{
	int i;
	int count=0;
	int port=7879;
	int ret;
	
	init_lights();
	sock=-1;
	
	while (sock<0 && port<9000) {
		sock=open_udp_server(port);
		port++;
	}

	if (sock<0) {
		return -1;
	}
	
	printf("sock: %i\n", sock);

	request_subscription(sock);
	ret=1;
	i=fcntl(sock,F_GETFL,0);
	fcntl(sock,F_SETFL,i|O_NONBLOCK);
	while (ret>=0) {
		if (read_data_packet(sock,light_data)<0) return -1;
		
//		for (i=0;i<60;i++) printf("%2X", light_data[i]);
//		printf("\n");
		
		ret=renderFrame(light_data);
//		printf("frame rendered.\n");
		count++;
		if ((count%250)==0) request_subscription(sock);
	}
	
	deinit_lights();
	return 0;
}

int read_data_packet(int sock, int *data)
{
	unsigned char net_data[NUM_SEGMENTS*6];
	int sizeaddr = sizeof(struct sockaddr);
	struct sockaddr_in from;
	int ret,i;
	
	ret=recvfrom(sock,&net_data,NUM_SEGMENTS*6,0,(struct sockaddr *) &from,&sizeaddr);
	
	if (ret<0) {
		if(errno==EAGAIN)
		    return 0;
		printf("%s\n",inet_ntoa(from.sin_addr));
		perror("read_data_packet: recvfrom");
		return -1;
	}
	
	if (ret<(NUM_SEGMENTS*6)) {
		fprintf(stderr,"Small packet of size: %i\n",ret);
		return -1;
	}
	
	for (i=0;i<60;i++) data[i]=net_data[i];
	
	return 0;
}

int request_subscription(int sock)
{
	struct sockaddr_in to;
	char b='A';
	
	to.sin_family=AF_INET;
	to.sin_addr.s_addr=inet_addr(SERVER_IP);
	to.sin_port=htons(SERVER_PORT);
	
	sendto(sock,&b,1,0,(struct sockaddr *) &to,sizeof(struct sockaddr));
	return 0;
}

int open_udp_server(int port)
{
	int sock;
	struct sockaddr_in my_addr;
	
	if ((sock=socket(AF_INET, SOCK_DGRAM, 0))==-1) {
		perror("start_net: creating socket");
		return -1;
	}
	
	my_addr.sin_family=AF_INET;
	my_addr.sin_addr.s_addr=INADDR_ANY;
	my_addr.sin_port=htons(port);
	bzero(&(my_addr.sin_zero),8);
	
	if (bind(sock, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) <0) {
		perror("start_net: binding socket");
		fprintf(stderr,"ERROR!\n");
		close(sock);
		return -1;
	}
	
	return sock;
}


