#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "../interface_2001.h"

/* this distributes the light feed out to any number of hosts who request
   it.  It keeps a linked list containing all the currently subscribed hosts.
   Every ten packets it walks the list and decrements the time to live.
   If after some fixed time a subscribe packet is not received, the listener
   is dropped from the list. */
   
struct subscribe_list {
	struct sockaddr_in host;
	int ttl;
	
	struct subscribe_list *next;
};


unsigned short server_port=7878;

static struct subscribe_list *root=NULL;

static int light_data[NUM_SEGMENTS*6];
static int sock=-1;
static struct sockaddr_in my_addr;

int send_frame();

int init_lights()
{
	sock=socket(AF_INET, SOCK_DGRAM, 0);
	
	if (sock<0) return -1;
	
	my_addr.sin_family=AF_INET;
	my_addr.sin_addr.s_addr=INADDR_ANY;
	my_addr.sin_port=htons(server_port);
	bzero(&my_addr.sin_zero,8);
	
	if (bind(sock, (struct sockaddr *) &my_addr, sizeof(struct sockaddr)) <0) {
		perror("could not bind to socket\n");
		close(sock);
		return -1;
	}
	
	fcntl(sock,F_SETFL,O_NONBLOCK);
	
	printf("Network Lights Client Initialized.\n");
	
	return 0;
}


int deinit_lights()
{
	if (sock>0) close(sock);
	return 0;
}

void light_state(int light, int state)
{
	int box,ns;
	
	box=light+1;
	ns=state*255;

	light_data[light*6]=ns; light_data[light*6+1]=ns;
	light_data[light*6+2]=ns; light_data[light*6+3]=ns;
	light_data[light*6+4]=ns; light_data[light*6+5]=ns;	

	send_frame();
}

int check_socket()
{
	char buf[1500];
	int ret;
	struct sockaddr_in from;
	struct subscribe_list *ptr;
	int flag;
	int sizeaddr = sizeof(struct sockaddr);
	
	do {
        ret=recvfrom(sock,&buf,1500,0,(struct sockaddr *) &from, &sizeaddr);

        if (ret>0) {
        	flag=-1;
        	// we received a packet... check to see if he is currently subscribed
        	for (ptr=root;ptr!=NULL;ptr=ptr->next) {
        		if ((memcmp(&ptr->host.sin_addr.s_addr,&from.sin_addr.s_addr,4)==0) &&
        			(memcmp(&ptr->host.sin_port,&from.sin_port,2)==0)) {
        			ptr->ttl=75;
        			flag=1;
        			break;
        		}
        	}
        	if (flag==-1) { // he was not in the list
        		ptr=malloc(sizeof(struct subscribe_list));
        		ptr->ttl=300;
        		memcpy(&ptr->host,&from,sizeof(struct sockaddr_in));
        		ptr->next=root;
        		root=ptr;
        	}
        }
	} while (ret>0);	
	return 0;
}

int cull_subscribers()
{
	struct subscribe_list *ptr,*prev_ptr,*oldptr;
	
	for (ptr=root,prev_ptr=NULL;ptr!=NULL;prev_ptr=ptr, ptr=ptr->next) {
//		printf("host: %s  ttl:%i\n",inet_ntoa(ptr->host.sin_addr),ptr->ttl);
		ptr->ttl--;
		if (ptr->ttl==0) { // this dude gets deleted
			if (prev_ptr) {
				prev_ptr->next=ptr->next;
				oldptr=ptr;
				ptr=ptr->next;
			} else {
				root=ptr->next;
				oldptr=ptr;
				ptr=ptr->next;
			}
			free(oldptr);
			if (ptr==NULL) break;
		}
	}
	return 0;
}

int renderFrame(int *data)
{
	// look through the data and write out any regpairs that differ from the last frame
	int i;
	static int count=0;
	
	count++;
	for (i=0;i<(NUM_SEGMENTS*6);i+=2) {
		if ((data[i]!=light_data[i]) || (data[i+1]!=light_data[i+1])) {
			light_data[i]=data[i]; light_data[i+1]=data[i+1];
		}
	}
	send_frame();
	check_socket();	// look for new listeners
	
	if ((count%10)==0) cull_subscribers();	// look for old subscribers

	return 0;
}

int send_frame()
{
	int ret,i;
	unsigned char net_data[NUM_SEGMENTS*6];
	struct subscribe_list *ptr;
	
	for (i=0;i<(NUM_SEGMENTS*6);i++) {
		net_data[i]=light_data[i];
		if (light_data[i]<0) net_data[i]=0;
		if (light_data[i]>255) net_data[i]=255;
	}
		
	for (ptr=root;ptr!=NULL;ptr=ptr->next) {	
		ret=sendto(sock,&net_data,NUM_SEGMENTS*6,0,
			(struct sockaddr *) &ptr->host,sizeof(struct sockaddr));
	}

	return 0;
}
