#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <termios.h>

/* Receives data for the lights over a configurable UDP port and writes
   it out to the light interface, be it the real lights or a simulation. */
   
#include "interface_2001.h"

#define SERVER_IP "131.151.188.33"
#define SERVER_PORT 7878

struct feed {
	char *desc;
	unsigned char *host;
	unsigned short port;
};

struct feed feed_list[] = {
	{"Josh's Main Music","131.151.188.33",7878},
	{"Test pattern","131.151.188.33",7879},
	{"Pong! Pong!","131.151.188.33", 7880},
	{"Brad 1","131.151.188.79",7878},
	{"Brad 2","131.151.188.79",7879},
	{"Brad 3","131.151.188.79", 7880},
	{"Josh's local comp", "10.0.0.13",7878},
	{NULL,NULL,0}};
	
int selected_feed=0;
int num_feeds=7;

int next_port=7881;
int sock;
static int light_data[NUM_SEGMENTS*6];
int scale=256;

int main(int argc, char **argv)
{
	int i;
	int count=0;
	int ret;
	int prevret;
   	fd_set fsr;
   	struct timeval tv;
   	int c,t;
	
	init_lights();
	
	sock=open_udp_server(next_port++);
	if (sock<0) {
		return -1;
	}
	printf("sock: %i\n", sock);

	request_subscription(sock);
	i=fcntl(sock,F_GETFL,0);
	fcntl(sock,F_SETFL,i|O_NONBLOCK);
	
	set_raw_mode();
		
	prevret=0;
	
	display_menu();
	
	for (;;) {
//		printf("%i  sock:%i\n",t++,sock);
    	FD_ZERO(&fsr);    	FD_SET(sock,&fsr);		FD_SET(0,&fsr);
    	tv.tv_sec=0;    	tv.tv_usec=500000;
    	ret=select(sock+1,&fsr,NULL,NULL,&tv);
    	if (ret<0) {
//    		perror("select");
    		continue;
    	}
    	    	
		if (ret!=prevret) {
			if (ret==0) {
				printf("\nFeed is not live!      \n");
			} else {
				printf("\nFeed is live!    %i    \n",ret);
			}
			fflush(stdout);
		}
    	prevret=ret;

		if (FD_ISSET(sock,&fsr)) {
//			printf("starting to write data\n");
			// we got light data
//			if (FD_ISSET(0,&fsr)) printf("we have a character\n");
			ret=read_data_packet(sock,light_data);
//			if (FD_ISSET(0,&fsr)) printf("we have a character1\n");
			scale_frame();
//			if (FD_ISSET(0,&fsr)) printf("we have a character2\n");
			if (ret>0) renderFrame(light_data);
//			printf("frame rendered\n");
		}
		
		if (FD_ISSET(0,&fsr)) {
//			printf("got a character\n");
			// we got a character on stdin
			c=getc(stdin);
			switch (c) {
			case 'q':
				exit(0);
			case 'u':
			case 'U':
				scale=scale+16;
				if (scale>256) scale=256;
				break;
			case 'd':
			case 'D':
				scale=scale-16;
				if (scale<0) scale=0;
				break;
			}
			if (c>='0' && c<='9') {
   				c=c-'0';
   				if (c<num_feeds) selected_feed=c;
   				close(sock);

            	sock=open_udp_server(next_port++);
            	if (next_port>10000) next_port=7879;

            	if (sock<0) {
            		return -1;
            	}
            	printf("sock: %i\n", sock);

            	request_subscription(sock);
            	i=fcntl(sock,F_GETFL,0);
            	fcntl(sock,F_SETFL,i|O_NONBLOCK);
			}	
				
			display_menu();
		}
		
//		printf("checking for subscript renewal\n");
		
		count++;
		if ((count%250)==0) {
//			printf("requesting subscription\n");
			request_subscription(sock);
//			printf("subscription requested\n");
		}
		
//		printf("end of loop\n");
		
	}
	
	deinit_lights();
	return 0;
}

int scale_frame()
{
	float a;
	float b;
	int i;
	
	a=(float)scale/256.0f;
	for	(i=0;i<(NUM_SEGMENTS*6);i++) {
		b=(float)light_data[i]*a;
		light_data[i]=(int) b;
	}

	return 0;
}

int set_raw_mode()
{
	 int fd = STDIN_FILENO;
	struct termios t;

	if (tcgetattr(fd, &t) < 0)
	{
		perror("tcgetattr");
		return -1;
	}

	t.c_lflag &= ~ICANON;

	if (tcsetattr(fd, TCSANOW, &t) < 0)
	{
		perror("tcsetattr");
		return -1;
	}

	setbuf(stdin, NULL);

	return 0;
}



int display_menu()
{
	int i;
	system("clear");
	printf("FEEDS\n-----\n");
	
	for (i=0;feed_list[i].host!=NULL;i++) {
		if (i==selected_feed) printf("=>"); 
			else printf("  ");
		printf("%i. %s \t\t %s:%i\n",i,feed_list[i].desc,
				feed_list[i].host,feed_list[i].port);
	}
	printf("\n");
	printf("Scale _U_p / _D_own : %i\n",scale);
	printf("\n");
}

int read_data_packet(int sock, int *data)
{
	unsigned char net_data[NUM_SEGMENTS*6];
	int sizeaddr = sizeof(struct sockaddr);
	struct sockaddr_in from;
	int ret,i;
	
//	printf("recvfrom(sock=%i,........)\n",sock);
	ret=recvfrom(sock,&net_data,NUM_SEGMENTS*6,0,(struct sockaddr *) &from,&sizeaddr);
	
	if (ret<0) {
		if (errno==EAGAIN)
			return 0;
		printf("%s\n",inet_ntoa(from.sin_addr));
		perror("read_data_packet: recvfrom");
		return -1;
	}
	
	if (ret<(NUM_SEGMENTS*6)) {
		fprintf(stderr,"Small packet of size: %i\n",ret);
		return -1;
	}
	
	for (i=0;i<60;i++) data[i]=net_data[i];
	
	return 1;
}

int request_subscription(int sock)
{
	struct sockaddr_in to;
	char b='A';
	
	to.sin_family=AF_INET;
	to.sin_addr.s_addr=inet_addr(feed_list[selected_feed].host);
	to.sin_port=htons(feed_list[selected_feed].port);
	
	sendto(sock,&b,1,0,(struct sockaddr *) &to,sizeof(struct sockaddr));
	return 0;
}

int open_udp_server(int port)
{
	int sock;
	struct sockaddr_in my_addr;
	
	if ((sock=socket(AF_INET, SOCK_DGRAM, 0))==-1) {
		perror("start_net: creating socket");
		return -1;
	}
	
	my_addr.sin_family=AF_INET;
	my_addr.sin_addr.s_addr=INADDR_ANY;
	my_addr.sin_port=htons(port);
	bzero(&(my_addr.sin_zero),8);
	
	if (bind(sock, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) <0) {
		perror("start_net: binding socket");
		fprintf(stderr,"ERROR!\n");
		close(sock);
		return -1;
	}
	
	return sock;
}


