#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "pattern.h"

#define PI 3.14159f
#define SCALE_DIV 50.0
#define SLOPE_MULT 10.0
#define FADE_STEP 48

double get_n_scaled_values(gint16 raw_data[2][256],double *lev,int n)
{
    static double scale=50.0,slope=2.0;
    double d,b,avg_height,s,pos;
    int i,j,last_bound,this_bound;
    b=pow(256,1.0/(double)(n+2));  // Throw away highest freq. data
    last_bound=0;
    avg_height=s=0.0;
    for(i=j=0;i<n;i++)
    {
        lev[i]=0.0;
        this_bound=pow(b,i+1);
        for(;j<this_bound;j++)
            lev[i]+=(double)raw_data[0][j]>0?raw_data[0][j]:-raw_data[0][j];
        pos=((double)i+0.5)/(double)n-0.5;
        lev[i]=slope*pos+log(0.6+scale*(double)lev[i]/32767.0/
          (double)(this_bound-last_bound));
        avg_height+=lev[i];
        s+=lev[i]*pos;
        if(lev[i]<0.0)
            lev[i]=0.0;
        else if(lev[i]>1.0)
            lev[i]=1.0;
    }
    avg_height/=(double)n;
    s/=(double)n;
    d=0.4-avg_height;
    scale*=exp(d/SCALE_DIV);
    slope-=s*s*s*SLOPE_MULT;
//    printf("scale = %f, slope = %f\r",scale,slope);
//    fflush(stdout);
    if(avg_height>=1.0)
        return 1.0;
    if(avg_height<=0.0)
        return 0.0;
    return avg_height;
}

void fade_hallway(gint light_data[NUM_SEGMENTS*6])
{
    int i;
    for(i=NUM_SEGMENTS*6-1;i>=0;i--)
    {
        light_data[i]-=FADE_STEP;
        if(light_data[i]<0)
            light_data[i]=0;
    }
}

static int se_center=0;	// 0 is the default center
						// 1 is from the left
						// 2 is from the right
static int se_fade=20;	// how quickly to fade bars

gint simple_equalizer_init(gchar *param)
{
	se_center=rand()%3;
	se_fade=rand()%20+10;
	printf("simple_equalizer_init called\n");
	return 0;
}

gint simple_equalizer_deinit(void)
{
	printf("simple_equalizer_deinit called\n");
	return 0;
}

gint simple_equalizer_frame(gdouble bar_averages[NUM_SEGMENTS],
							gint16 bar_heights[NUM_SEGMENTS],
							gint light_data[NUM_SEGMENTS*6], 
							gint16 raw_data[2][256], gint time_left)
{
	int i,x,t;

	if (se_center==0) {
       for(i = 0; i < NUM_SEGMENTS; i++) {
           light_data[i*6]=light_data[i*6+5]=max(min((bar_heights[i]-10)*5 ,255) ,0);
           light_data[i*6+1]=light_data[i*6+4]=max(min(bar_heights[i]*6-165 ,255) ,0);
           light_data[i*6+2]=light_data[i*6+3]=max(min(bar_heights[i]*5-245 ,255) ,0);
       }
    } else if (se_center>0) {
    	for (i=0;i<NUM_SEGMENTS;i++) {
    		for (x=0;x<6;x++) {
    			if (se_center==1) t=x; else t=5-x;
    			if ((bar_heights[i]-10)>(x*16)) 
    				light_data[i*6+t]=max(((bar_heights[i]-10)-(x*16))*16,255);
    			else
    				light_data[i*6+t]=min(light_data[i*6+t]-se_fade,0);
    		}
    	}
    }	

	return 0;
}

float sw_offset;
float sw_dense;
float sw_offset_var;
float sw_speed;
int sw_dir;

gint sine_wave_init(gchar *param)
{
	sw_offset=0;
	sw_dense=rand()%4+4;
	sw_offset_var=rand()%20/20.0f;
	sw_speed=((rand()%15)+3)/10.0f;
	sw_dir=rand()%2*2-1;
	printf("sine_wave_init called  dense=%f  offset_var=%f  speed=%f\n",
		sw_dense,sw_offset_var,sw_speed);
	return 0;
}

gint sine_wave_deinit(void)
{
	printf("sine_wave_deinit called\n");
	return 0;
}

gint sine_wave_frame(gdouble bar_averages[NUM_SEGMENTS],
							gint16 bar_heights[NUM_SEGMENTS],
							gint light_data[NUM_SEGMENTS*6], 
							gint16 raw_data[2][256], gint time_left)
{
	float x1,x2,val;
	float t_dense;
	int y,ival;
	int x;
    double totval;

    get_n_scaled_values(raw_data,&totval,1);
	
	for(y=0;y<NUM_SEGMENTS;y++) {
		x1=y;
		x2=2*PI*sw_speed/30.0;
		val=sin(x1*2.0*PI/sw_dense+x2*sw_offset*(float) sw_dir)*127.0+127.0f;
		ival=val;
		for (x=0;x<6;x++) light_data[y*6+x]=ival;
	}
	sw_offset++;
	sw_offset+=(totval-0.5)*sw_offset_var;
	return 0;
}


static gint re_direction=0,re_fade;

gint ring_equalizer_init(gchar *param)
{
    re_direction=0;
    re_fade=-1;
    if (param)
    {
        parse_string(param,"i","dir",&re_direction);
        parse_string(param,"i","fade",&re_fade);
    }
    if(!re_direction)
    {
        re_direction=(rand()&1)*2-1;
        printf("Randomly picking direction %d.\n",re_direction);
    }
    else
        printf("Being told to use direction %d.\n",re_direction);
    if(re_fade<0)
    {
        re_fade=rand()&15;
        printf("Randomly picking fade val %d.\n",re_fade);
    }
    return 0;
}

gint ring_equalizer_deinit(void)
{
	printf("ring_equalizer_deinit called\n");
	return 0;
}

gint ring_equalizer_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    int i,j;
    double vals[6];
    get_n_scaled_values(raw_data,vals,6);
    if(re_direction>0)
    {
        for(i=NUM_SEGMENTS*6-1;i>=6;i--)
        {
            light_data[i]=light_data[i-6]-re_fade;
            if(light_data[i]<0)
                light_data[i]=0;
        }
        i=0;
    }
    else
        for(i=0;i<(NUM_SEGMENTS-1)*6;i++)
        {
            light_data[i]=light_data[i+6]-re_fade;
            if(light_data[i]<0)
                light_data[i]=0;
        }
    for(j=0;j<6;j++,i++)
        light_data[i]=(int)(255.0*vals[j]);
    return 0;
}


static int tl_max_lines=12;
static int tl_fade_count=10;
static int tl_new_line_chance=3;
static int tl_move_count=2;
static int tl_dir=1;

static int *line_position=NULL;
static int *line_count=NULL;
static int num_lines;

gint trailing_lines_init(gchar *param)
{
	int i;

	printf("trailing_lines_init called\n");

	parse_string(param,"iiiii",
			"max_lines",&tl_max_lines,
			"fade_count",&tl_fade_count,
			"new_line_chance",&tl_new_line_chance,
			"move_count",&tl_move_count,
			"dir",&tl_dir);

	line_position=(int *) malloc(sizeof(int)*tl_max_lines);
	line_count=(int *) malloc(sizeof(int)*tl_max_lines);
	num_lines=0;
	
	for (i=0;i<tl_max_lines;i++) {
		line_position[i]=-1;
		line_count[i]=-1;
	}
	
	return 0;
}

gint trailing_lines_deinit(void)
{
	if (line_position) free(line_position); line_position=NULL;
	if (line_count) free(line_count); line_count=NULL;
	return 0;
}

gint trailing_lines_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
	int i;
	
	for (i=0;i<60;i++) {
		light_data[i]-=tl_fade_count;
		if (light_data[i]<0) light_data[i]=0;
	}
	
     // now go through and move all the lines that might need to be
     
     for (i=0;i<tl_max_lines;i++) {
         if (line_count[i]>0) {
             line_count[i]--;  
             if (line_count[i]==0) {
                 line_count[i]=tl_move_count;
                 line_position[i]+=(6*tl_dir);
             }
         }
     }
      
     // check for out of range line positions
     for (i=0;i<tl_max_lines;i++) {
         if ((line_position[i]>=(NUM_SEGMENTS*6)) || (line_position[i]<0)) {
             line_position[i]=-1;
             line_count[i]=-1;   
             num_lines--;
         }
     }

     if ( (( (rand()%tl_new_line_chance)==0)  && (num_lines<tl_max_lines)) || (num_lines==0)) {
         // create a new line
         
         for (i=0;i<tl_max_lines;i++) {
             if (line_position[i]==-1) break;
         }
          
         line_position[i]=rand()%6 + ( (tl_dir>0) ? 0 : (NUM_SEGMENTS-1)*6);
         line_count[i]=tl_move_count;
     }


	// set all the line heads to 255
	for (i=0;i<tl_max_lines;i++) {  
	    if (line_position[i]>=0) {  
	        light_data[line_position[i]]=255;
	    }
	}
	return 0;
}


static int rise_center=5;
static int rise_speed=3;
static int rise_diff=20;
static int rise_bot=0;
static int rise_mid=-40;
static int rise_top=-80;

static int rise_seg_level[NUM_SEGMENTS];

gint rise_init(gchar *param)
{
	int y;
	printf("rise_init called\n");
	
	parse_string(param,"iiiiii",
		"center",&rise_center,
		"speed",&rise_speed,
		"diff",&rise_diff,
		"bot",&rise_bot,
		"mid",&rise_mid,
		"top",&rise_top);
		
	rise_seg_level[rise_center]=0;
	for (y=(rise_center+1);y<NUM_SEGMENTS;y++) rise_seg_level[y]=rise_seg_level[y-1]-rise_diff;
	for (y=(rise_center-1);y>=0;y--) rise_seg_level[y]=rise_seg_level[y+1]-rise_diff;
	
	return 0;
}

gint rise_deinit()
{
	return 0;
}

gint rise_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
	int y;
	int flag=0;

	memset(light_data,0,sizeof(int)*NUM_SEGMENTS*6);

	
	for (y=0;y<NUM_SEGMENTS;y++) {
		rise_seg_level[y]+=rise_speed;
		
		if ((rise_seg_level[y]+rise_bot)>0)
			light_data[y*6]=light_data[y*6+5]=min(rise_seg_level[y]+rise_bot,255);
		if ((rise_seg_level[y]+rise_mid)>0)
			light_data[y*6+1]=light_data[y*6+4]=min(rise_seg_level[y]+rise_mid,255);
		if ((rise_seg_level[y]+rise_top)>0)
			light_data[y*6+2]=light_data[y*6+3]=min(rise_seg_level[y]+rise_top,255);
			
		if ((rise_seg_level[y]+rise_top)<255) flag=1;
	}
	
	if (flag==0) return -1;
	return 0;
}

static int bar_equalizer_dir=1;

gint bar_equalizer_init(gchar *param)
{
    printf("bar_equalizer_init called\n");
    if (param)
    {
        parse_string(param,"i","dir",&bar_equalizer_dir);
        printf("Being told to use direction %d.\n",bar_equalizer_dir);
    }
    else
    {
        bar_equalizer_dir=(rand()%2)*2-1;
        printf("Randomly picking direction %d.\n",bar_equalizer_dir);
    }
    return 0;
}

gint bar_equalizer_deinit()
{
    return 0;
}

gint bar_equalizer_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    int i,j,k,len;
    double val;
    get_n_scaled_values(raw_data,&val,1);
    fade_hallway(light_data);
    len=(int)(val*(double)NUM_SEGMENTS);
    for(i=0,k=bar_equalizer_dir>0?0:NUM_SEGMENTS-1;i<len;i++,
     k+=bar_equalizer_dir)
        for(j=0;j<6;j++)
            light_data[k*6+j]=255;
    return 0;
}


gint random_spots_init(gchar *param)
{

	return 0;
}

gint random_spots_deinit()
{

	return 0;
}

gint random_spots_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
	int i;
	int f;
	double val;
	get_n_scaled_values(raw_data,&val,1);
	for (i=val*60.0;i>=0;i--) {
		f=rand()%60;
		if (light_data[f]>0) light_data[f]=0;
			else light_data[f]=255;
	}
	
	return 0;
}

gint center_meter_init(gchar *param)
{
    return 0;
}

gint center_meter_deinit()
{
    return 0;
}

gint center_meter_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    int t,i,j,len;
    double val;
    get_n_scaled_values(raw_data,&val,1);
    fade_hallway(light_data);
    len=(int)(val*(double)(NUM_SEGMENTS>>1));
    for(i=0;i<len;i++)
        for(j=0;j<6;j++)
        {
            light_data[(i+(NUM_SEGMENTS>>1))*6+j]=255;
            light_data[((NUM_SEGMENTS>>1)-1-i)*6+j]=255;
        }
    return 0;
}

gint brightness_level_init(gchar *param)
{
    return 0;
}

gint brightness_level_deinit()
{
    return 0;
}

gint brightness_level_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    int i,l;
    double val;
    get_n_scaled_values(raw_data,&val,1);
    l=(int)(255.0*val);
    for(i=0;i<NUM_SEGMENTS*6;i++)
    {
        light_data[i]-=16;
        if(l>light_data[i])
            light_data[i]=l;
        if(light_data[i]<0)
            light_data[i]=0;
    }
    return 0;
}

gint vertical_meter_init(gchar *param)
{
    return 0;
}

gint vertical_meter_deinit()
{
    return 0;
}

gint vertical_meter_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    int i,j,k,len;
    double val;
    get_n_scaled_values(raw_data,&val,1);
    fade_hallway(light_data);
    len=(int)(val*3.0);
    for(i=0;i<len;i++)
        for(j=0;j<NUM_SEGMENTS;j++)
            light_data[j*6+i]=light_data[j*6+5-i]=255;
    return 0;
}

static int sm_direction,sm_fade;

gint swirl_meter_init(gchar *param)
{
    sm_direction=0;
    sm_fade=-1;
    if (param)
    {
        parse_string(param,"i","dir",&sm_direction);
        parse_string(param,"i","fade",&sm_fade);
    }
    if(!sm_direction)
    {
        sm_direction=(rand()&1)*2-1;
        printf("Randomly picking direction %d.\n",sm_direction);
    }
    else
        printf("Being told to use direction %d.\n",sm_direction);
    if(sm_fade<0)
    {
        sm_fade=rand()&7;
        printf("Randomly picking fade val %d.\n",sm_fade);
    }
    return 0;
}

gint swirl_meter_deinit(void)
{
	return 0;
}

gint swirl_meter_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    double vals[2],d;
    static double rot=0.0;
    int i,j;
    get_n_scaled_values(raw_data,vals,2);
    if(sm_direction>0)
    {
        for(i=NUM_SEGMENTS*6-1;i>=6;i--)
        {
            light_data[i]=light_data[i-6]-sm_fade;
            if(light_data[i]<0)
                light_data[i]=0;
        }
        i=0;
    }
    else
        for(i=0;i<(NUM_SEGMENTS-1)*6;i++)
        {
            light_data[i]=light_data[i+6]-sm_fade;
            if(light_data[i]<0)
                light_data[i]=0;
        }
    for(j=0;j<6;j++,i++)
    {
        d=rot-(double)j;
        if(d<0.0)
            d=-d;
        if(d>4.0)
            d=8.0-d;
        if(d>=3.4)
            light_data[i]=255.0*vals[1];
        else if(d<=0.4)
            light_data[i]=255.0*vals[0];
        else
            light_data[i]=0;
    }
    rot+=0.2;
    while(rot>=8.0)
        rot-=8.0;
    return 0;
}

gint vs_fade;

gint vertical_spectrum_init(gchar *param)
{
    vs_fade=-1;
    if (param)
        parse_string(param,"i","fade",&re_fade);
    if(vs_fade<0)
    {
        vs_fade=rand()&127;
        printf("Randomly picking fade val %d.\n",vs_fade);
    }
    return 0;
}

gint vertical_spectrum_deinit()
{
    return 0;
}

gint vertical_spectrum_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    int i,j,t;
    double vals[NUM_SEGMENTS];
    get_n_scaled_values(raw_data,vals,NUM_SEGMENTS);
    for(i=0;i<NUM_SEGMENTS*6;i+=6)
    {
        light_data[i+2]=light_data[i+1]-vs_fade;
        if(light_data[i+2]<0)
            light_data[i+2]=0;
        light_data[i+3]=light_data[i+4]-vs_fade;
        if(light_data[i+3]<0)
            light_data[i+3]=0;
        light_data[i+1]=light_data[i]-vs_fade;
        if(light_data[i+1]<0)
            light_data[i+1]=0;
        light_data[i+4]=light_data[i+5]-vs_fade;
        if(light_data[i+4]<0)
            light_data[i+4]=0;
        t=256.0*vals[i/6];
        light_data[i]-=8;
        if(t>light_data[i])
            light_data[i]=t;
        light_data[i+5]-=8;
        if(t>light_data[i+5])
            light_data[i+5]=t;
    }
    return 0;
}

static int ss_rate,ss_invert;

gint spinning_stripe_init(gchar *param)
{
    ss_rate=0;
    ss_invert=-1;
    if (param)
    {
        parse_string(param,"i","rate",&ss_rate);
        parse_string(param,"i","invert",&ss_invert);
    }
    if(ss_rate==0)
    {
        ss_rate=(rand()&63)+32;
        if(rand()&1)
            ss_rate=-ss_rate;
        printf("Randomly picking rate %d.\n",ss_rate);
    }
    if(ss_invert<0)
    {
        ss_invert=rand()&1;
        printf("Randomly picking invert %d.\n",ss_invert);
    }
    return 0;
}

gint spinning_stripe_deinit(void)
{
	return 0;
}

gint spinning_stripe_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    double d;
    static double rot=0.0;
    int i,j,t;
    for(j=0;j<6;j++)
    {
        d=rot-(double)j;
        if(d<0.0)
            d=-d;
        if(d>4.0)
            d=8.0-d;
        if((d>=3.5)||(d<=0.5))
            for(i=j;i<NUM_SEGMENTS*6;i+=6)
                light_data[i]=ss_invert?0:255;
        else
            if(ss_invert)
                for(i=j;i<NUM_SEGMENTS*6;i+=6)
                {
                    light_data[i]+=80;
                    if(light_data[i]>255)
                        light_data[i]=255;
                }
            else
                for(i=j;i<NUM_SEGMENTS*6;i+=6)
                {
                    light_data[i]-=80;
                    if(light_data[i]<0)
                        light_data[i]=0;
                }
    }
    rot+=(double)ss_rate/256.0;
    while(rot>=8.0)
        rot-=8.0;
    while(rot<0.0)
        rot+=8.0;
    return 0;
}

int wc_delay,wc_pos,wc_count;

gint warp_core_init(gchar *param)
{
    wc_delay=0;
    if(param)
        parse_string(param,"i","delay",&wc_delay);
    if(!wc_delay)
        wc_delay=rand()&3;
    return 0;
}

gint warp_core_deinit()
{
    return 0;
}

gint warp_core_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    int t;
    if(wc_count>=wc_delay)
    {
        wc_count=0;
        wc_pos++;
        if(wc_pos==(NUM_SEGMENTS>>1))
        {
            t=wc_pos*6;
            wc_pos=0;
        }
    }
    else
        wc_count++;
    fade_hallway(light_data);
    t=wc_pos*6;
    light_data[t]=light_data[t+1]=light_data[t+2]=
     light_data[t+3]=light_data[t+4]=light_data[t+5]=255;
    t=(NUM_SEGMENTS-wc_pos-1)*6;
    light_data[t]=light_data[t+1]=light_data[t+2]=
     light_data[t+3]=light_data[t+4]=light_data[t+5]=255;
    return 0;
}

double exp_rad=0;

gint explosion_init(gchar *param)
{
    exp_rad=0.0;
    return 0;
}

gint explosion_deinit()
{
    return 0;
}

gint explosion_frame(gdouble bar_averages[NUM_SEGMENTS],
                            gint16 bar_heights[NUM_SEGMENTS],
                            gint light_data[NUM_SEGMENTS*6],
                            gint16 raw_data[2][256], gint time_left)
{
    int i,j,t;
    double x,y;
    if(exp_rad>=0)
    {
        for(i=0;i<NUM_SEGMENTS;i++)
        {
            x=(double)i-(double)(NUM_SEGMENTS-1)/2.0;
            for(j=0;j<6;j++)
            {
                y=(double)(j<<1)-5.0;
                if(sqrt(x*x+y*y)<exp_rad)
                    light_data[i*6+j]=255;
            }
        }
        exp_rad+=.5;
        if(exp_rad>7.0)
            exp_rad=-1.0;
    }
    else
    {
        for(i=0;i<NUM_SEGMENTS;i++)
        {
            x=(double)i-(double)(NUM_SEGMENTS-1)/2.0;
            for(j=0;j<6;j++)
            {
                y=(double)(j<<1)-5.0;
                if(sqrt(x*x+y*y)<-exp_rad)
                {
                    t=i*6+j;
                    light_data[t]-=32;
                    if(light_data[t]<0)
                        light_data[t]=0;
                }
            }
        }
        exp_rad-=.5;
        if(exp_rad<-15.0)
            return -1;
    }
    return 0;
}
